=== modified file 'bzrlib/tests/HTTPTestUtil.py'
--- bzrlib/tests/HTTPTestUtil.py	2007-03-14 08:12:20 +0000
+++ bzrlib/tests/HTTPTestUtil.py	2007-03-14 16:39:24 +0000
@@ -17,6 +17,7 @@
 from cStringIO import StringIO
 import errno
 from SimpleHTTPServer import SimpleHTTPRequestHandler
+import re
 import socket
 import urlparse
 
@@ -232,14 +233,17 @@
         """Redirect a single HTTP request to another host"""
         valid = TestingHTTPRequestHandler.parse_request(self)
         if valid:
-            rhost, rport = self.server.test_case_server.redirected_to_address()
-            if rhost is not None and rport is not None:
-                # Just redirect permanently
-                self.send_response(301)
-                target = 'http://%s:%s%s' % (rhost, rport, self.path)
+            tcs = self.server.test_case_server
+            code, target = tcs.is_redirected(self.path)
+            if code is not None and target is not None:
+                # Redirect as instructed
+                self.send_response(code)
                 self.send_header('Location', target)
                 self.end_headers()
                 return False # The job is done
+            else:
+                # We leave the parent class serve the request
+                pass
         return valid
 
 
@@ -248,15 +252,38 @@
 
     def __init__(self, request_handler=RedirectRequestHandler):
         HttpServer.__init__(self, request_handler)
-        self._redirect_to_host = None
-        self._redirect_to_port = None
-
-    def redirect_to(self, redir_host, redir_port):
-        self._redirect_to_host = redir_host
-        self._redirect_to_port = redir_port
-
-    def redirected_to_address(self):
-        return self._redirect_to_host, self._redirect_to_port
+        # redirections is a list of tuples (source, target, code)
+        # - source is a regexp for the paths requested
+        # - target is a replacement for re.sub describing where
+        #   the request will be redirected
+        # - code is the http error code associated to the
+        #   redirection (301 permanent, 302 temporarry, etc
+        self.redirections = []
+
+    def redirect_to(self, host, port):
+        """Redirect all requests to a specific host:port"""
+        self.redirections = [('(.*)',
+                              r'http://%s:%s\1' % (host, port) ,
+                              301)]
+
+    def is_redirected(self, path):
+        """Is the path redirected by this server.
+
+        :param path: the requested relative path
+
+        :returns: a tuple (code, target) if a matching
+             redirection is found, (None, None) otherwise.
+        """
+        code = None
+        target = None
+        for (rsource, rtarget, rcode) in self.redirections:
+            target, match = re.subn(rsource, rtarget, path)
+            if match:
+                code = rcode
+                break # The first match wins
+            else:
+                target = None
+        return code, target
 
 
 class TestCaseWithRedirectedWebserver(TestCaseWithTwoWebservers):

=== modified file 'bzrlib/tests/HttpServer.py'
--- bzrlib/tests/HttpServer.py	2007-03-14 08:12:20 +0000
+++ bzrlib/tests/HttpServer.py	2007-03-14 16:39:24 +0000
@@ -254,8 +254,9 @@
         BaseHTTPServer.HTTPServer.__init__(self, server_address,
                                            RequestHandlerClass)
         # test_case_server can be used to communicate between the
-        # tests and the server, allowing dynamic behaviors to be
-        # defined from the tests cases.
+        # tests and the server (or the request handler and the
+        # server), allowing dynamic behaviors to be defined from
+        # the tests cases.
         self.test_case_server = test_case_server
 
 

=== modified file 'bzrlib/tests/test_http.py'
--- bzrlib/tests/test_http.py	2007-03-13 17:00:20 +0000
+++ bzrlib/tests/test_http.py	2007-03-14 16:39:24 +0000
@@ -62,6 +62,7 @@
 from bzrlib.transport.http import (
     extract_auth,
     HttpTransportBase,
+    _urllib2_wrappers,
     )
 from bzrlib.transport.http._urllib import HttpTransport_urllib
 
@@ -953,13 +954,6 @@
                         TestCaseWithWebserver):
     """Test the Range header in GET methods for pycurl implementation"""
 
-# FIXME: http implementations do not redirect silently anymore
-# (they do not redirect at all in fact). The mechanism is still
-# in place at the _urllib2_wrappers.Request level and tests
-# should be written to exercise it. For the pycurl implementation
-# the redirection have been deleted as we may deprecate pycurl
-# and I have no place to keep a working implementation.
-#  -- vila 20070212
 
 class TestHTTPRedirections(object):
     """Test redirection between http servers.
@@ -1016,6 +1010,92 @@
     """Tests redirections for pycurl implementation"""
 
 
+class RedirectedRequest(_urllib2_wrappers.Request):
+    """Request following redirections"""
+
+    init_orig = _urllib2_wrappers.Request.__init__
+
+    def __init__(self, method, url, *args, **kwargs):
+        RedirectedRequest.init_orig(self, method, url, args, kwargs)
+        self.follow_redirections = True
+
+
+class TestHTTPSilentRedirections_urllib(TestCaseWithRedirectedWebserver):
+    """Test redirections provided by urllib.
+
+    http implementations do not redirect silently anymore (they
+    do not redirect at all in fact). The mechanism is still in
+    place at the _urllib2_wrappers.Request level and these tests
+    exercise it.
+
+    For the pycurl implementation
+    the redirection have been deleted as we may deprecate pycurl
+    and I have no place to keep a working implementation.
+    -- vila 20070212
+    """
+
+    _transport = HttpTransport_urllib
+
+    def setUp(self):
+        super(TestHTTPSilentRedirections_urllib, self).setUp()
+        self.setup_redirected_request()
+        self.addCleanup(self.cleanup_redirected_request)
+        self.build_tree_contents([('a','a'),
+                                  ('1/',),
+                                  ('1/a', 'redirected once'),
+                                  ('2/',),
+                                  ('2/a', 'redirected twice'),
+                                  ('3/',),
+                                  ('3/a', 'redirected thrice'),
+                                  ('4/',),
+                                  ('4/a', 'redirected 4 times'),
+                                  ('5/',),
+                                  ('5/a', 'redirected 5 times'),
+                                  ],)
+
+        self.old_transport = self._transport(self.old_server.get_url())
+
+    def setup_redirected_request(self):
+        self.original_class = _urllib2_wrappers.Request
+        _urllib2_wrappers.Request = RedirectedRequest
+
+    def cleanup_redirected_request(self):
+        _urllib2_wrappers.Request = self.original_class
+
+    def create_transport_secondary_server(self):
+        """Create the secondary server, redirections are defined in the tests"""
+        return HTTPServerRedirecting()
+
+    def test_one_redirection(self):
+        t = self.old_transport
+
+        req = RedirectedRequest('GET', t.abspath('a'))
+        req.follow_redirections = True
+        new_prefix = 'http://%s:%s' % (self.new_server.host,
+                                       self.new_server.port)
+        self.old_server.redirections = \
+            [('(.*)', r'%s/1\1' % (new_prefix), 301),]
+        self.assertEquals('redirected once',t._perform(req).read())
+
+    def test_five_redirections(self):
+        t = self.old_transport
+
+        req = RedirectedRequest('GET', t.abspath('a'))
+        req.follow_redirections = True
+        old_prefix = 'http://%s:%s' % (self.old_server.host,
+                                       self.old_server.port)
+        new_prefix = 'http://%s:%s' % (self.new_server.host,
+                                       self.new_server.port)
+        self.old_server.redirections = \
+            [('/1(.*)', r'%s/2\1' % (old_prefix), 302),
+             ('/2(.*)', r'%s/3\1' % (old_prefix), 303),
+             ('/3(.*)', r'%s/4\1' % (old_prefix), 307),
+             ('/4(.*)', r'%s/5\1' % (new_prefix), 301),
+             ('(/[^/]+)', r'%s/1\1' % (old_prefix), 301),
+             ]
+        self.assertEquals('redirected 5 times',t._perform(req).read())
+
+
 class TestDoCatchRedirections(TestCaseWithRedirectedWebserver):
     """Test transport.do_catching_redirections.
 

=== modified file 'bzrlib/transport/http/_urllib2_wrappers.py'
--- bzrlib/transport/http/_urllib2_wrappers.py	2007-03-13 17:00:20 +0000
+++ bzrlib/transport/http/_urllib2_wrappers.py	2007-03-14 16:39:24 +0000
@@ -503,9 +503,9 @@
         # of creating a new one, but the urllib2.Request object
         # has a too complicated creation process to provide a
         # simple enough equivalent update process. Instead, when
-        # redirecting, we only update the original request with a
-        # reference to the following request in the redirect
-        # chain.
+        # redirecting, we only update the following request in
+        # the redirect chain with a reference to the parent
+        # request .
 
         # Some codes make no sense in our context and are treated
         # as errors:
@@ -542,7 +542,7 @@
         else:
             raise urllib2.HTTPError(req.get_full_url(), code, msg, headers, fp)
 
-    def http_error_30x(self, req, fp, code, msg, headers):
+    def http_error_302(self, req, fp, code, msg, headers):
         """Requests the redirected to URI.
 
         Copied from urllib2 to be able to fake_close the
@@ -596,24 +596,7 @@
 
         return self.parent.open(redirected_req)
 
-    http_error_302 = http_error_303 = http_error_307 = http_error_30x
-
-    def http_error_301(self, req, fp, code, msg, headers):
-        response = self.http_error_30x(req, fp, code, msg, headers)
-        # If one or several 301 responses occur during the
-        # redirection chain, we MUST update the original request
-        # to indicate where the URI where finally found.
-
-        original_req = req
-        while original_req.parent is not None:
-            original_req = original_req.parent
-        if original_req.redirected_to is None:
-            # Only the last occurring 301 (the deepest in the
-            # recursive call chain) should be taken into
-            # account i.e. the first occurring here when
-            # redirected_to has not yet been set.
-            original_req.redirected_to = req.redirected_to
-        return response
+    http_error_301 = http_error_303 = http_error_307 = http_error_302
 
 
 class ProxyHandler(urllib2.ProxyHandler):

